@@ -1,51 +1,227 @@
 /*
  * IceWM
  *
- * Copyright (C) 1997 Marko Macek
+ * Copyright (C) 1997,1998 Marko Macek
  */
+#include "config.h"
+#include "yfull.h"
+#include "ykey.h"
+#include "wmframe.h"
 
-#include "icewm.h"
+#include "wmclient.h"
+#include "wmstatus.h"
+#include "yapp.h"
+
+void YFrameWindow::snapTo(int &wx, int &wy,
+                          int rx1, int ry1, int rx2, int ry2,
+                          int &flags)
+{
+    int d = snapDistance;
+
+    if (flags & 4) { // snap to container window (root, workarea)
+        int iw = width();
+        if (flags & 8)
+            iw -= 2 * borderX();
+
+        if (flags & 1) { // x
+            int wxw = wx + iw;
+
+            if (wx >= rx1 - d && wx <= rx1 + d) {
+                wx = rx1;
+                flags &= ~1;
+            } else if (wxw >= rx2 - d && wxw <= rx2 + d) {
+                wx = rx2 - iw;
+                flags &= ~1;
+            }
+        }
+    } else { // snap to other window
+        int iw = width();
+        int ih = height();
+        bool touchY = false;
+
+        //if (wy >= ry1 && wy <= ry2 ||
+        //    wy + ih >= rx1 && wy + ih <= ry2)
+        if (wy <= ry2 && wy + ih >= ry1)
+            touchY = true;
+
+        if ((flags & 1) && touchY) { // x
+            int wxw = wx + iw;
+
+            if (wx >= rx2 - d && wx <= rx2 + d) {
+                wx = rx2;
+                flags &= ~1;
+            } else if (wxw >= rx1 - d && wxw <= rx1 + d) {
+                wx = rx1 - iw;
+                flags &= ~1;
+            }
+        }
+    }
+    if (flags & 4) {
+        int ih = height();
+
+        if (flags & 16)
+            ih -= 2 * borderY();
+
+        if (flags & 2) { // y
+            int wyh = wy + ih;
+
+            if (wy >= ry1 - d && wy <= ry1 + d) {
+                wy = ry1;
+                flags &= ~2;
+            } else if (wyh >= ry2 - d && wyh <= ry2 + d) {
+                wy = ry2 - ih;
+                flags &= ~2;
+            }
+        }
+    } else {
+        int iw = width();
+        int ih = height();
+        bool touchX = false;
+
+        //if (wx >= rx1 && wx <= rx2 ||
+        //    wx + iw >= rx1 && wx + iw <= rx2)
+        if (wx <= rx2 && wx + iw >= rx1)
+            touchX = true;
+
+        if ((flags & 2) && touchX) { // y
+            int wyh = wy + ih;
+
+            if (wy >= ry2 - d && wy <= ry2 + d) {
+                wy = ry2;
+                flags &= ~2;
+            } else if (wyh >= ry1 - d && wyh <= ry1 + d) {
+                wy = ry1 - ih;
+                flags &= ~2;
+            }
+        }
+    }
+}
+
+void YFrameWindow::snapTo(int &wx, int &wy) {
+    YFrameWindow *f = manager->topLayer();
+    int flags = 1 | 2;
+    int xp = wx, yp = wy;
+    int rx1, ry1, rx2, ry2;
+
+    /// !!! clean this up, it should snap to the closest thing it finds
+    
+    // try snapping to the border first
+    flags |= 4;
+    if (xp < manager->minX(getLayer()) || xp + int(width()) > manager->maxX(getLayer())) {
+        xp += borderX();
+        flags |= 8;
+    }
+    if (yp < manager->minY(getLayer()) || yp + int(height()) > manager->maxY(getLayer())) {
+        yp += borderY();
+        flags |= 16;
+    }
+    snapTo(xp, yp,
+           manager->minX(getLayer()),
+           manager->minY(getLayer()),
+           manager->maxX(getLayer()),
+           manager->maxY(getLayer()),
+           flags);
+    if (flags & 8) {
+        xp -= borderX();
+        flags &= ~8;
+    }
+    if (flags & 16) {
+        yp -= borderY();
+        flags &= ~16;
+    }
+    if (xp < 0 || xp + width() > desktop->width()) {
+        xp += borderX();
+        flags |= 8;
+    }
+    if (yp < 0 || yp + height() > desktop->height()) {
+        yp += borderY();
+        flags |= 16;
+    }
+    snapTo(xp, yp, 0, 0, manager->width(), manager->height(), flags);
+    if (flags & 8) {
+        xp -= borderX();
+        flags &= ~8;
+    }
+    if (flags & 16) {
+        yp -= borderY();
+        flags &= ~16;
+    }
+    flags &= ~4;
+
+    if (flags & (1 | 2)) {
+        // we only snap to windows below, hope that's ok
+        while (f) {
+            if (f != this && f->visible()) {
+                rx1 = f->x();
+                ry1 = f->y();
+                rx2 = f->x() + f->width();
+                ry2 = f->y() + f->height();
+                snapTo(xp, yp, rx1, ry1, rx2, ry2, flags);
+                if (!(flags & (1 | 2)))
+                    break;
+            }
+            f = f->nextLayer();
+        }
+    }
+    wx = xp;
+    wy = yp;
+}
 
 void YFrameWindow::drawOutline(int x, int y, int w, int h) {
     int bw = (wsBorderX + wsBorderY) / 2;
-    
+    static GC outlineGC = None;
+
+    if (outlineGC == None) {
+        XGCValues gcv;
+
+        gcv.foreground = YColor(clrActiveBorder).pixel();
+        gcv.function = GXxor;
+        gcv.graphics_exposures = False;
+        gcv.line_width = (wsBorderX + wsBorderY) / 2;
+        gcv.subwindow_mode = IncludeInferiors;
+
+        outlineGC = XCreateGC(app->display(), desktop->handle(),
+                              GCForeground |
+                              GCFunction |
+                              GCGraphicsExposures |
+                              GCLineWidth |
+                              GCSubwindowMode,
+                              &gcv);
+    }
+
     x += bw / 2;
     y += bw / 2;
     w -= bw;
     h -= bw;
-    XDrawRectangle(app->display(),
-                   app->root()->handle(),
-                   outlineGC,
+    XDrawRectangle(app->display(), manager->handle(), outlineGC,
                    x, y, w, h);
 }
 
 int YFrameWindow::handleMoveKeys(const XKeyEvent &key, int &newX, int &newY) {
-    int k = XKeycodeToKeysym(app->display(), key.keycode, 0);
+    KeySym k = XKeycodeToKeysym(app->display(), key.keycode, 0);
     int m = KEY_MODMASK(key.state);
     int factor = 1;
 
-    if (m == ControlMask)
-        factor = 16;
-    else if (m == ShiftMask)
+    if (m == ShiftMask)
         factor = 4;
-    
-    if (k == XK_Left)
+
+    if (k == XK_Left || k == XK_KP_Left)
         newX -= factor;
-    else if (k == XK_Right)
+    else if (k == XK_Right || k == XK_KP_Right)
         newX += factor;
-    else if (k == XK_Up)
+    else if (k == XK_Up || k == XK_KP_Up)
         newY -= factor;
-    else if (k == XK_Down)
+    else if (k == XK_Down || k == XK_KP_Down)
         newY += factor;
-    else if (k == XK_Home)
-        newX = 0 - borderX();
-    else if (k == XK_End)
-        newX = app->root()->width() - width() + borderX();
-    else if (k == XK_Prior)
-        newY = 0 - borderY();
-    else if (k == XK_Next)
-        newY = app->root()->height() - height() + borderY();
-    else if (k == XK_Return)
+    else if (k == XK_Home || k == XK_KP_Home)
+        newX = manager->minX(getLayer()) - borderX();
+    else if (k == XK_End || k == XK_KP_End)
+        newX = manager->maxX(getLayer()) - width() + borderX();
+    else if (k == XK_Prior || k == XK_KP_Prior)
+        newY = manager->minY(getLayer()) - borderY();
+    else if (k == XK_Next || k == XK_KP_Next)
+        newY = manager->maxY(getLayer()) - height() + borderY();
+    else if (k == XK_Return || k == XK_KP_Enter)
         return -1;
     else if (k ==  XK_Escape) {
         newX = origX;
@@ -55,21 +231,19 @@ int YFrameWindow::handleMoveKeys(const XKeyEvent &key, int &newX, int &newY) {
         return 0;
     return 1;
 }
-            
+
 int YFrameWindow::handleResizeKeys(const XKeyEvent &key,
                                    int &newX, int &newY, int &newWidth, int &newHeight,
                                    int incX, int incY)
 {
-    int k = XKeycodeToKeysym(app->display(), key.keycode, 0);
+    KeySym k = XKeycodeToKeysym(app->display(), key.keycode, 0);
     int m = KEY_MODMASK(key.state);
     int factor = 1;
 
-    if (m == ControlMask)
-        factor = 16;
-    else if (m == ShiftMask)
+    if (m == ShiftMask)
         factor = 4;
-    
-    if (k == XK_Left) {
+
+    if (k == XK_Left || k == XK_KP_Left) {
         if (grabX == 0) {
             grabX = -1;
         } else if (grabX == 1) {
@@ -78,7 +252,7 @@ int YFrameWindow::handleResizeKeys(const XKeyEvent &key,
             newWidth += incX * factor;
             newX -= incX * factor;
         }
-    } else if (k == XK_Right) {
+    } else if (k == XK_Right || k == XK_KP_Right) {
         if (grabX == 0) {
             grabX = 1;
         } else if (grabX == 1) {
@@ -87,7 +261,7 @@ int YFrameWindow::handleResizeKeys(const XKeyEvent &key,
             newWidth -= incX * factor;
             newX += incX * factor;
         }
-    } else if (k == XK_Up) {
+    } else if (k == XK_Up || k == XK_KP_Up) {
         if (grabY == 0) {
             grabY = -1;
         } else if (grabY == 1) {
@@ -96,7 +270,7 @@ int YFrameWindow::handleResizeKeys(const XKeyEvent &key,
             newHeight += incY * factor;
             newY -= incY * factor;
         }
-    } else if (k == XK_Down) {
+    } else if (k == XK_Down || k == XK_KP_Down) {
         if (grabY == 0) {
             grabY = 1;
         } else if (grabY == 1) {
@@ -105,7 +279,7 @@ int YFrameWindow::handleResizeKeys(const XKeyEvent &key,
             newHeight -= incY * factor;
             newY += incY * factor;
         }
-    } else if (k == XK_Return)
+    } else if (k == XK_Return || k == XK_KP_Enter)
         return -1;
     else if (k ==  XK_Escape) {
         newX = origX;
@@ -121,37 +295,51 @@ int YFrameWindow::handleResizeKeys(const XKeyEvent &key,
 void YFrameWindow::handleMoveMouse(const XMotionEvent &motion, int &newX, int &newY) {
     int mouseX = motion.x_root;
     int mouseY = motion.y_root;
-    
+
     constrainMouseToWorkspace(mouseX, mouseY);
-    
+
     newX = mouseX - buttonDownX;
     newY = mouseY - buttonDownY;
-    
+
     constrainPositionByModifier(newX, newY, motion);
 
     newX += borderX();
     newY += borderY();
     int n = -2;
-    if (newX + int(width() + n * borderX()) > app->maxX())
-        if (newX + int(width() + n * borderX()) < int(app->maxX() + EdgeResistance))
-            newX = app->maxX() - width() - n * borderX();
-//        else
-//            newX -= EdgeResistance;
-    if (newY + int(height() + n * borderY()) > app->maxY())
-        if (newY + int(height() + n * borderY()) < int(app->maxY() + EdgeResistance))
-            newY = app->maxY() - height() - n * borderY();
-//        else
-//            newY -= EdgeResistance;
-    if (newX < app->minX())
-        if (newX > int(- EdgeResistance + app->minX()))
-            newX = app->minX();
-//        else
-//            newX += EdgeResistance;
-    if (newY < app->minY())
-        if (newY > int(- EdgeResistance + app->minY()))
-            newY = app->minY();
-//        else
-//            newY += EdgeResistance;
+
+    if (!(motion.state & app->AltMask)) {
+        if (EdgeResistance == 10000) {
+            if (newX + int(width() + n * borderX()) > manager->maxX(getLayer()))
+                newX = manager->maxX(getLayer()) - width() - n * borderX();
+            if (newY + int(height() + n * borderY()) > manager->maxY(getLayer()))
+                newY = manager->maxY(getLayer()) - height() - n * borderY();
+            if (newX < manager->minX(getLayer()))
+                newX = manager->minX(getLayer());
+            if (newY < manager->minY(getLayer()))
+                newY = manager->minY(getLayer());
+        } else if (/*EdgeResistance >= 0 && %%% */ EdgeResistance < 10000) {
+            if (newX + int(width() + n * borderX()) > manager->maxX(getLayer()))
+                if (newX + int(width() + n * borderX()) < int(manager->maxX(getLayer()) + EdgeResistance))
+                    newX = manager->maxX(getLayer()) - width() - n * borderX();
+                else if (motion.state & ShiftMask)
+                    newX -= EdgeResistance;
+            if (newY + int(height() + n * borderY()) > manager->maxY(getLayer()))
+                if (newY + int(height() + n * borderY()) < int(manager->maxY(getLayer()) + EdgeResistance))
+                    newY = manager->maxY(getLayer()) - height() - n * borderY();
+                else if (motion.state & ShiftMask)
+                    newY -= EdgeResistance;
+            if (newX < manager->minX(getLayer()))
+                if (newX > int(- EdgeResistance + manager->minX(getLayer())))
+                    newX = manager->minX(getLayer());
+                else if (motion.state & ShiftMask)
+                    newX += EdgeResistance;
+            if (newY < manager->minY(getLayer()))
+                if (newY > int(- EdgeResistance + manager->minY(getLayer())))
+                    newY = manager->minY(getLayer());
+                else if (motion.state & ShiftMask)
+                    newY += EdgeResistance;
+        }
+    }
     newX -= borderX();
     newY -= borderY();
 }
@@ -161,23 +349,23 @@ void YFrameWindow::handleResizeMouse(const XMotionEvent &motion,
 {
     int mouseX = motion.x_root;
     int mouseY = motion.y_root;
-    
+
     constrainMouseToWorkspace(mouseX, mouseY);
-    
+
     if (grabX == -1) {
         newX = mouseX - buttonDownX;
         newWidth = width() + x() - newX;
     } else if (grabX == 1) {
         newWidth = mouseX + buttonDownX - x();
     }
-    
+
     if (grabY == -1) {
         newY = mouseY - buttonDownY;
         newHeight = height() + y() - newY;
     } else if (grabY == 1) {
         newHeight = mouseY + buttonDownY - y();
     }
-    
+
     newWidth -= 2 * borderX();
     newHeight -= 2 * borderY() + titleY();
     client()->constrainSize(newWidth, newHeight,
@@ -186,7 +374,7 @@ void YFrameWindow::handleResizeMouse(const XMotionEvent &motion,
                             (grabY ? YFrameClient::csKeepY : 0));
     newWidth += 2 * borderX();
     newHeight += 2 * borderY() + titleY();
-    
+
     if (grabX == -1)
         newX = x() + width() - newWidth;
     if (grabY == -1)
@@ -196,13 +384,90 @@ void YFrameWindow::handleResizeMouse(const XMotionEvent &motion,
 
 void YFrameWindow::outlineMove() {
     int xx = x(), yy = y();
+    unsigned int modifiers = 0;
 
     XGrabServer(app->display());
-    XSync(app->display(), 0);
+    XSync(app->display(), False);
     drawOutline(xx, yy, width(), height());
     while (1) {
         XEvent xev;
-        
+
+        XWindowEvent(app->display(),
+                     handle(),
+                     KeyPressMask | ExposureMask |
+                     ButtonPressMask | ButtonReleaseMask | PointerMotionMask, &xev);
+        switch (xev.type) {
+        case KeyPress:
+            modifiers = xev.xkey.state;
+            {
+                int ox = xx;
+                int oy = yy;
+                int r;
+
+                switch (r = handleMoveKeys(xev.xkey, xx, yy)) {
+                case 0:
+                    break;
+                case 1:
+                case -2:
+                    if (xx != ox || yy != oy) {
+                        drawOutline(ox, oy, width(), height());
+#ifndef LITE
+                        statusMoveSize->setStatus(this, xx, yy, width(), height());
+#endif
+                        drawOutline(xx, yy, width(), height());
+                    }
+                    if (r == -2)
+                        goto end;
+                    break;
+                case -1:
+                    goto end;
+                }
+            }
+            break;
+        case ButtonPress:
+        case ButtonRelease:
+            modifiers = xev.xbutton.state;
+            goto end;
+        case MotionNotify:
+            {
+                int ox = xx;
+                int oy = yy;
+
+                handleMoveMouse(xev.xmotion, xx, yy);
+                if (xx != ox || yy != oy) {
+#ifndef LITE
+                    statusMoveSize->setStatus(this, xx, yy, width(), height());
+#endif
+                    drawOutline(ox, oy, width(), height());
+                    drawOutline(xx, yy, width(), height());
+                }
+            }
+            break;
+        }
+    }
+end:
+    drawOutline(xx, yy, width(), height());
+    XSync(app->display(), False);
+    moveWindow(xx, yy);
+    XUngrabServer(app->display());
+}
+
+void YFrameWindow::outlineResize() {
+    int xx = x(), yy = y(), ww = width(), hh = height();
+    int incX = 1;
+    int incY = 1;
+
+    if (client()->sizeHints()) {
+        incX = client()->sizeHints()->width_inc;
+        incY = client()->sizeHints()->height_inc;
+    }
+
+    XGrabServer(app->display());
+    XSync(app->display(), False);
+    drawOutline(xx, yy, ww, hh);
+    while (1) {
+        XEvent xev;
+
         XWindowEvent(app->display(),
                      handle(),
                      KeyPressMask | ExposureMask |
@@ -212,16 +477,110 @@ void YFrameWindow::outlineMove() {
             {
                 int ox = xx;
                 int oy = yy;
-                
-                switch (handleMoveKeys(xev.xkey, xx, yy)) {
+                int ow = ww;
+                int oh = hh;
+                int r;
+
+                switch (r = handleResizeKeys(xev.xkey, xx, yy, ww, hh, incX, incY)) {
+                case 0:
+                    break;
+                case -2:
+                case 1:
+                    if (ox != xx || oy != yy || ow != ww || oh != hh) {
+                        drawOutline(ox, oy, ow, oh);
+#ifndef LITE
+                        statusMoveSize->setStatus(this, xx, yy, ww, hh);
+#endif
+                        drawOutline(xx, yy, ww, hh);
+                    }
+                    if (r == -2)
+                        goto end;
+                    break;
+                case -1:
+                    goto end;
+                }
+            }
+            break;
+        case ButtonPress:
+        case ButtonRelease:
+            goto end;
+        case MotionNotify:
+            {
+                int ox = xx;
+                int oy = yy;
+                int ow = ww;
+                int oh = hh;
+
+                handleResizeMouse(xev.xmotion, xx, yy, ww,hh);
+                if (ox != xx || oy != yy || ow != ww || oh != hh) {
+                    drawOutline(ox, oy, ow, oh);
+#ifndef LITE
+                    statusMoveSize->setStatus(this, xx, yy, ww, hh);
+#endif
+                    drawOutline(xx, yy, ww, hh);
+                }
+            }
+            break;
+        }
+    }
+end:
+    drawOutline(xx, yy, ww, hh);
+    XSync(app->display(), False);
+    setGeometry(xx, yy, ww, hh);
+    XUngrabServer(app->display());
+}
+
+void YFrameWindow::manualPlace() {
+    int xx = x(), yy = y();
+    Cursor grabPointer = movePointer;
+
+    grabX = borderX();
+    grabY = borderY();
+    origX = x();
+    origY = y();
+    origW = width();
+    origH = height();
+    buttonDownX = 0;
+    buttonDownY = 0;
+
+    if (!app->grabEvents(desktop, grabPointer,
+                         ButtonPressMask |
+                         ButtonReleaseMask |
+                         PointerMotionMask))
+        return ;
+    XGrabServer(app->display());
+#ifndef LITE
+    statusMoveSize->begin(this);
+#endif
+    drawOutline(xx, yy, width(), height());
+    while (1) {
+        XEvent xev;
+
+        XMaskEvent(app->display(),
+                   KeyPressMask |
+                   ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
+                   &xev);
+        switch (xev.type) {
+        case KeyPress:
+            {
+                int ox = xx;
+                int oy = yy;
+                int r;
+
+                switch (r = handleMoveKeys(xev.xkey, xx, yy)) {
                 case 0:
                     break;
                 case 1:
+                case -2:
                     if (xx != ox || yy != oy) {
                         drawOutline(ox, oy, width(), height());
+#ifndef LITE
                         statusMoveSize->setStatus(this, xx, yy, width(), height());
+#endif
                         drawOutline(xx, yy, width(), height());
                     }
+                    if (r == -2)
+                        goto end;
                     break;
                 case -1:
                     goto end;
@@ -238,7 +597,9 @@ void YFrameWindow::outlineMove() {
 
                 handleMoveMouse(xev.xmotion, xx, yy);
                 if (xx != ox || yy != oy) {
+#ifndef LITE
                     statusMoveSize->setStatus(this, xx, yy, width(), height());
+#endif
                     drawOutline(ox, oy, width(), height());
                     drawOutline(xx, yy, width(), height());
                 }
@@ -248,82 +609,15 @@ void YFrameWindow::outlineMove() {
     }
 end:
     drawOutline(xx, yy, width(), height());
-    XSync(app->display(), 0);
+#ifndef LITE
+    statusMoveSize->end();
+#endif
     moveWindow(xx, yy);
+    app->releaseEvents();
     XUngrabServer(app->display());
 }
 
-void YFrameWindow::outlineResize() {
-    int xx = x(), yy = y(), ww = width(), hh = height();
-    int incX = 1;
-    int incY = 1;
-    
-    if (client()->sizeHints()) {
-        incX = client()->sizeHints()->width_inc;
-        incY = client()->sizeHints()->height_inc;
-    }
-
-    XGrabServer(app->display());
-    XSync(app->display(), 0);
-    drawOutline(xx, yy, ww, hh);
-    while (1) {
-        XEvent xev;
-        
-        XWindowEvent(app->display(),
-                     handle(),
-                     KeyPressMask | ExposureMask |
-                     ButtonPressMask | ButtonReleaseMask | PointerMotionMask, &xev);
-        switch (xev.type) {
-        case KeyPress:
-            {
-                int ox = xx;
-                int oy = yy;
-                int ow = ww;
-                int oh = hh;
-                
-                switch (handleResizeKeys(xev.xkey, xx, yy, ww, hh, incX, incY)) {
-                case 0:
-                    break;
-                case 1:
-                    if (ox != xx || oy != yy || ow != ww || oh != hh) {
-                        drawOutline(ox, oy, ow, oh);
-                        statusMoveSize->setStatus(this, xx, yy, ww, hh);
-                        drawOutline(xx, yy, ww, hh);
-                    }
-                    break;
-                case -1:
-                    goto end;
-                }
-            }
-            break;
-        case ButtonPress:
-        case ButtonRelease:
-            goto end;
-        case MotionNotify:
-            {
-                int ox = xx;
-                int oy = yy;
-                int ow = ww;
-                int oh = hh;
-                
-                handleResizeMouse(xev.xmotion, xx, yy, ww,hh);
-                if (ox != xx || oy != yy || ow != ww || oh != hh) {
-                    drawOutline(ox, oy, ow, oh);
-                    statusMoveSize->setStatus(this, xx, yy, ww, hh);
-                    drawOutline(xx, yy, ww, hh);
-                }
-            }
-            break;
-        }
-    }
-end:
-    drawOutline(xx, yy, ww, hh);
-    XSync(app->display(), 0);
-    setGeometry(xx, yy, ww, hh);
-    XUngrabServer(app->display());
-}
-
-void YFrameWindow::handleKey(const XKeyEvent &key) {
+bool YFrameWindow::handleKey(const XKeyEvent &key) {
     if (key.type == KeyPress) {
         if (movingWindow) {
             int newX = x();
@@ -337,7 +631,7 @@ void YFrameWindow::handleKey(const XKeyEvent &key) {
                 endMoveSize();
                 break;
             case 0:
-                return ;
+                return true;
             case 1:
                 moveWindow(newX, newY);
                 break;
@@ -367,7 +661,7 @@ void YFrameWindow::handleKey(const XKeyEvent &key) {
                                         (grabY ? YFrameClient::csKeepY : 0));
                 newWidth += 2 * borderX();
                 newHeight += 2 * borderY() + titleY();
-                
+
                 if (grabX == -1)
                     newX = x() + width() - newWidth;
                 if (grabY == -1)
@@ -377,8 +671,10 @@ void YFrameWindow::handleKey(const XKeyEvent &key) {
                             newY,
                             newWidth,
                             newHeight);
-                
+
+#ifndef LITE
                 statusMoveSize->setStatus(this);
+#endif
                 break;
             case -2:
                 setGeometry(newX, newY, newWidth, newHeight);
@@ -387,94 +683,107 @@ void YFrameWindow::handleKey(const XKeyEvent &key) {
                 endMoveSize();
                 break;
             }
-        } else {
-            int k = XKeycodeToKeysym(app->display(), key.keycode, 0);
-            int m = KEY_MODMASK(key.state);
-            
-            if (m == Mod1Mask + ShiftMask) {
-                if (k == XK_F6)
-                    wmPrevWindow();
-                else if (k == XK_Tab)
-                    switchWindow->begin(0);
-            } else if (m == Mod1Mask) {
-                if (k == XK_Tab)
-                    switchWindow->begin(1);
-                else if (k == ' ')
-                    popupSystemMenu();
-                else if (k == XK_F2)
-                    wmOccupyAllOrCurrent();
-                else if (k == XK_F3)
-                    wmLower();
-                else if (k == XK_F4)
-                    wmClose();
-                else if (k == XK_F5)
+        } else if (app->AltMask != 0) {
+            KeySym k = XKeycodeToKeysym(app->display(), key.keycode, 0);
+            unsigned int m = KEY_MODMASK(key.state);
+            unsigned int vm = VMod(m);
+
+            if (!isRollup() && !isIconic() &&
+                key.window != handle())
+                return true;
+
+            if (IS_WMKEY(k, vm, gKeyWinClose)) {
+                if (canClose()) wmClose();
+            } else if (IS_WMKEY(k, vm, gKeyWinPrev)) {
+                wmPrevWindow();
+            } else if (IS_WMKEY(k, vm, gKeyWinMaximizeVert)) {
+                wmMaximizeVert();
+            } else if (IS_WMKEY(k, vm, gKeyWinRaise)) {
+                if (canRaise()) wmRaise();
+            } else if (IS_WMKEY(k, vm, gKeyWinOccupyAll)) {
+                wmOccupyAllOrCurrent();
+            } else if (IS_WMKEY(k, vm, gKeyWinLower)) {
+                if (canLower()) wmLower();
+            } else if (IS_WMKEY(k, vm, gKeyWinRestore)) {
+                wmRestore();
+            } else if (IS_WMKEY(k, vm, gKeyWinNext)) {
+                wmNextWindow();
+            } else if (IS_WMKEY(k, vm, gKeyWinMove)) {
+                if (canMove()) wmMove();
+            } else if (IS_WMKEY(k, vm, gKeyWinSize)) {
+                if (canSize()) wmSize();
+            } else if (IS_WMKEY(k, vm, gKeyWinMinimize)) {
+                if (canMinimize()) wmMinimize();
+            } else if (IS_WMKEY(k, vm, gKeyWinMaximize)) {
+                if (canMaximize()) wmMaximize();
+            } else if (IS_WMKEY(k, vm, gKeyWinHide)) {
+                if (canHide()) wmHide();
+            } else if (IS_WMKEY(k, vm, gKeyWinRollup)) {
+                if (canRollup()) wmRollup();
+            } else if (IS_WMKEY(k, vm, gKeyWinMenu)) {
+                popupSystemMenu();
+            }
+            if (isIconic() || isRollup()) {
+                if (k == XK_Return || k == XK_KP_Enter) {
                     wmRestore();
-                else if (k == XK_F6)
-                    wmNextWindow();
-                else if (k == XK_F7)
-                    wmMove();
-                else if (k == XK_F8)
-                    wmSize();
-                else if (k == XK_F9)
-                    wmMinimize();
-                else if (k == XK_F10)
-                    wmMaximize();
-                else if (k == XK_F11)
-                    wmHide();
-                else if (k == XK_F12)
-                    wmShade();
+                } else if ((k == XK_Menu) || (k == XK_F10 && m == ShiftMask)) {
+                    popupSystemMenu();
+                }
             }
         }
     }
+    return true;
 }
 
 void YFrameWindow::constrainPositionByModifier(int &x, int &y, const XMotionEvent &motion) {
-    int mask = motion.state & (ShiftMask | Mod1Mask);
+    unsigned int mask = motion.state & (ShiftMask | ControlMask);
 
     x += borderX();
     y += borderY();
     if (mask == ShiftMask) {
         x = x / 4 * 4;
         y = y / 4 * 4;
-    } else if (mask == Mod1Mask) {
-        x = x / 8 * 8;
-        y = y / 8 * 8;
     }
     x -= borderX();
     y -= borderY();
+
+    if (snapMove && !(mask & (ControlMask | ShiftMask))) {
+        snapTo(x, y);
+    }
 }
 
 void YFrameWindow::constrainMouseToWorkspace(int &x, int &y) {
-    if (x < app->minX())
-        x = app->minX();
-    else if (x >= app->maxX())
-        x = app->maxX() - 1;
-    if (y < app->minY())
-        y = app->minY();
-    else if (y >= app->maxY())
-        y = app->maxY() - 1;
+    if (x < manager->minX(getLayer()))
+        x = manager->minX(getLayer());
+    else if (x >= manager->maxX(getLayer()))
+        x = manager->maxX(getLayer()) - 1;
+    if (y < manager->minY(getLayer()))
+        y = manager->minY(getLayer());
+    else if (y >= manager->maxY(getLayer()))
+        y = manager->maxY(getLayer()) - 1;
 }
 
-int YFrameWindow::canSize() {
-    if (style & fsShaded)
-        return 0;
-    if (maximized() && !sizeMaximized)
-        return 0;
+bool YFrameWindow::canSize(bool horiz, bool vert) {
+    if (isRollup())
+        return false;
 #ifndef NO_MWM_HINTS
     if (!(client()->mwmFunctions() & MWM_FUNC_RESIZE))
-        return 0;
+        return false;
 #endif
-    return 1;
+    if (!sizeMaximized) {
+        if ((!vert || isMaximizedVert()) &&
+            (!horiz || isMaximizedHoriz()))
+            return false;
+    }
+    return true;
 }
 
-int YFrameWindow::canMove() {
-    //if (maximized() && !moveSizeMaximized)
-    //    return 0;
+bool YFrameWindow::canMove() {
 #ifndef NO_MWM_HINTS
     if (!(client()->mwmFunctions() & MWM_FUNC_MOVE))
-        return 0;
+        return false;
 #endif
-    return 1;
+    return true;
 }
 
 void YFrameWindow::startMoveSize(int doMove, int byMouse,
@@ -498,7 +807,7 @@ void YFrameWindow::startMoveSize(int doMove, int byMouse,
         grabPointer = movePointer;
     } else if (!doMove) {
         Cursor ptr = leftPointer;
-        
+
         if (grabY == -1) {
             if (grabX == -1)
                 ptr = sizeTopLeftPointer;
@@ -519,21 +828,21 @@ void YFrameWindow::startMoveSize(int doMove, int byMouse,
             else if (grabX == 1)
                 ptr = sizeRightPointer;
         }
-        
+
         if (grabX == 1)
             buttonDownX = x() + width() - mouseXroot;
         else if (grabX == -1)
             buttonDownX = mouseXroot - x();
-        
+
         if (grabY == 1)
             buttonDownY = y() + height() - mouseYroot;
         else if (grabY == -1)
             buttonDownY = mouseYroot - y();
-        
+
         grabPointer = ptr;
     }
 
-    XSync(app->display(), 0);
+    XSync(app->display(), False);
     if (!app->grabEvents(this, grabPointer,
                          ButtonPressMask |
                          ButtonReleaseMask |
@@ -547,7 +856,9 @@ void YFrameWindow::startMoveSize(int doMove, int byMouse,
     else
         sizingWindow = 1;
 
+#ifndef LITE
     statusMoveSize->begin(this);
+#endif
     if (doMove && !opaqueMove) {
         outlineMove();
         endMoveSize();
@@ -559,11 +870,13 @@ void YFrameWindow::startMoveSize(int doMove, int byMouse,
 
 void YFrameWindow::endMoveSize() {
     app->releaseEvents();
+#ifndef LITE
     statusMoveSize->end();
+#endif
     movingWindow = 0;
     sizingWindow = 0;
 }
-        
+
 void YFrameWindow::handleBeginDrag(const XButtonEvent &down, const XMotionEvent &motion) {
     if ((down.button == 3) && canMove()) {
         startMoveSize(1, 1,
@@ -573,49 +886,50 @@ void YFrameWindow::handleBeginDrag(const XButtonEvent &down, const XMotionEvent
     } else if ((down.button == 1) && canSize()) {
         grabX = 0;
         grabY = 0;
-        
+
         if (down.x < int(borderX())) grabX = -1;
         else if (width() - down.x <= borderX()) grabX = 1;
-        
+
         if (down.y < int(borderY())) grabY = -1;
         else if (height() - down.y <= borderY()) grabY = 1;
-        
+
         if (grabY != 0 && grabX == 0) {
             if (down.x < int(wsCornerX)) grabX = -1;
             else if (width() - down.x <= wsCornerX) grabX = 1;
         }
-        
+
         if (grabX != 0 && grabY == 0) {
             if (down.y < int(wsCornerY)) grabY = -1;
             else if (height() - down.y <= wsCornerY) grabY = 1;
         }
-        
+
         if (grabX != 0 || grabY != 0) {
             startMoveSize(0, 1,
                           grabX, grabY,
                           down.x_root, down.y_root);
-            
+
         }
     }
 }
 
 void YFrameWindow::moveWindow(int newX, int newY) {
+    if (newX >= int(manager->maxX(getLayer()) - borderX()))
+        newX = manager->maxX(getLayer()) - borderX();
 
-    if (newX >= int(app->maxX() - borderX()))
-        newX = app->maxX() - borderX();
+    if (newY >= int(manager->maxY(getLayer()) - borderY() - titleY()))
+        newY = manager->maxY(getLayer()) - borderY() - titleY();
 
-    if (newY >= int(app->maxY() - borderY() - titleY()))
-        newY = app->maxY() - borderY() - titleY();
-    
-    if (newX < int(app->minX() - (width() - borderX())))
-        newX = app->minX() - (width() - borderX());
+    if (newX < int(manager->minX(getLayer()) - (width() - borderX())))
+        newX = manager->minX(getLayer()) - (width() - borderX());
 
-    if (newY < int(app->minY() - (height() - borderY())))
-        newY = app->minY() - (height() - borderY());
+    if (newY < int(manager->minY(getLayer()) - (height() - borderY())))
+        newY = manager->minY(getLayer()) - (height() - borderY());
 
     setPosition(newX, newY);
-    
+
+#ifndef LITE
     statusMoveSize->setStatus(this);
+#endif
 }
 
 void YFrameWindow::handleDrag(const XButtonEvent &/*down*/, const XMotionEvent &/*motion*/) {
@@ -626,7 +940,7 @@ void YFrameWindow::handleEndDrag(const XButtonEvent &/*down*/, const XButtonEven
 
 void YFrameWindow::handleButton(const XButtonEvent &button) {
     if (button.type == ButtonPress) {
-        if (!(button.state & ControlMask)) {
+        if (button.button == 1 && !(button.state & ControlMask)) {
             activate();
             if (raiseOnClickFrame)
                 wmRaise();
@@ -651,7 +965,9 @@ void YFrameWindow::handleMotion(const XMotionEvent &motion) {
                     newY,
                     newWidth,
                     newHeight);
+#ifndef LITE
         statusMoveSize->setStatus(this);
+#endif
         return ;
     } else if (movingWindow) {
         int newX = x();
@@ -663,3 +979,4 @@ void YFrameWindow::handleMotion(const XMotionEvent &motion) {
     }
     YWindow::handleMotion(motion);
 }
+
